@@ -31,26 +31,28 @@ from urllib.error import (URLError, HTTPError)
 from urllib.parse import urlparse
 from urllib.request import urlopen
 import codecs
+import functools
 import json
 import os.path
 import re
-import shutil
 import subprocess
 import sys
 
+from abc import (ABC, abstractmethod)
+
 from debian.changelog import Changelog
 import debian.deb822
-import httplib2
 
 from contextlib import closing
 
 from ubuntutools.config import UDTConfig
-from ubuntutools.lp.lpapicache import (Launchpad, Distribution, PersonTeam,
+from ubuntutools.lp.lpapicache import (Launchpad, Distribution, PersonTeam, Project,
                                        SourcePackagePublishingHistory)
 from ubuntutools.lp.udtexceptions import (PackageNotFoundException,
                                           SeriesNotFoundException,
+                                          PocketDoesNotExistError,
                                           InvalidDistroValueError)
-from ubuntutools.misc import (download, verify_file_checksum)
+from ubuntutools.misc import (download, verify_file_checksum, verify_file_checksums)
 from ubuntutools.version import Version
 
 import logging
@@ -111,13 +113,20 @@ class Dsc(debian.deb822.Dsc):
         return True
 
 
-class SourcePackage(object):
+class SourcePackage(ABC):
     """Base class for source package downloading.
     Use DebianSourcePackage or UbuntuSourcePackage instead of using this
     directly.
     """
-    distribution = None
-    spph_class = SourcePackagePublishingHistory
+
+    @property
+    @abstractmethod
+    def distribution(self):
+        return None
+
+    @property
+    def spph_class(self):
+        return SourcePackagePublishingHistory
 
     def __init__(self, package=None, version=None, component=None,
                  *args, **kwargs):
@@ -129,7 +138,6 @@ class SourcePackage(object):
         latest version from the development series.
         """
         dscfile = kwargs.get('dscfile')
-        lp = kwargs.get('lp')
         mirrors = kwargs.get('mirrors', ())
         workdir = kwargs.get('workdir', '.')
         series = kwargs.get('series')
@@ -140,8 +148,14 @@ class SourcePackage(object):
 
         assert (package is not None or dscfile is not None)
 
+        if 'lp' in kwargs:
+            # deprecated - please don't use this; if you want to use an
+            # existing lp object, just call login_existing() directly
+            Logger.warning("Deprecation warning: please don't pass 'lp' to SourcePackage")
+            if not Launchpad.logged_in:
+                Launchpad.login_existing(kwargs['lp'])
+
         self.source = package
-        self._lp = lp
         self.binary = None
         self.try_binary = try_binary
         self.workdir = workdir
@@ -159,14 +173,22 @@ class SourcePackage(object):
         self._version = Version(version) if version else None
 
         # Mirrors
-        self.mirrors = list(mirrors)
-        if self.distribution:
-            self.masters = [UDTConfig.defaults['%s_MIRROR'
-                                               % self.distribution.upper()]]
+        self.mirrors = list(filter(None, mirrors))
+        self.masters = list(filter(None,
+                                   [UDTConfig.defaults.get(f'{self.distribution.upper()}_{suffix}')
+                                    for suffix in ["MIRROR", "PORTS_MIRROR", "INTERNAL_MIRROR"]]))
 
-        # if a dsc was specified, pull it to get the source/version info
+        # If provided a dscfile, process it now to set our source and version
         if self._dsc_source:
-            self.pull_dsc()
+            filename = os.path.basename(urlparse(self._dsc_source).path)
+            dst = os.path.join(self.workdir, filename)
+            download(self._dsc_source, dst)
+            with open(dst, 'rb') as f:
+                self._dsc = Dsc(f.read())
+            self.source = self._dsc['Source']
+            self._version = Version(self._dsc['Version'])
+            if self._verify_signature:
+                self._check_dsc_signature()
 
     @property
     def lp_spph(self):
@@ -174,12 +196,6 @@ class SourcePackage(object):
         if self._spph:
             return self._spph
 
-        if not Launchpad.logged_in:
-            if self._lp:
-                Launchpad.login_existing(self._lp)
-            else:
-                Launchpad.login_anonymously()
-
         archive = self.getArchive()
         params = {}
         if self._version:
@@ -265,15 +281,18 @@ class SourcePackage(object):
 
     @property
     def dsc(self):
-        "Return a the Dsc"
-        if self._dsc is None:
+        "Return the Dsc"
+        if not self._dsc:
             self.pull_dsc()
+            with open(self.dsc_pathname, 'rb') as f:
+                self._dsc = Dsc(f.read())
+            if self._verify_signature:
+                self._check_dsc_signature()
         return self._dsc
 
     def getDistribution(self):
         if not self._distribution:
             self._distribution = Distribution(self.distribution)
-
         return self._distribution
 
     def getArchive(self):
@@ -288,172 +307,114 @@ class SourcePackage(object):
         return os.path.join(mirror, 'pool', self.component, group,
                             self.source, filename)
 
+    def _archive_servers(self):
+        "Generator for mirror and master servers"
+        # Always provide the mirrors first
+        for server in self.mirrors:
+            yield server
+        # Don't repeat servers that are in both mirrors and masters
+        for server in set(self.masters) - set(self.mirrors):
+            yield server
+
     def _source_urls(self, name):
         "Generator of sources for name"
         if self._dsc_source:
             yield os.path.join(os.path.dirname(self._dsc_source), name)
-        for mirror in self.mirrors:
-            yield self._mirror_url(mirror, name)
-        for mirror in self.masters:
-            if mirror not in self.mirrors:
-                yield self._mirror_url(mirror, name)
+        for server in self._archive_servers():
+            yield self._mirror_url(server, name)
         if self.lp_spph.sourceFileUrl(name):
             yield self.lp_spph.sourceFileUrl(name)
 
     def _binary_urls(self, name, bpph):
         "Generator of URLs for name"
-        for mirror in self.mirrors:
-            yield self._mirror_url(mirror, name)
-        for mirror in self.masters:
-            if mirror not in self.mirrors:
-                yield self._mirror_url(mirror, name)
+        for server in self._archive_servers():
+            yield self._mirror_url(server, name)
         if bpph.binaryFileUrl(name):
             yield bpph.binaryFileUrl(name)
         if bpph.getUrl():
             yield bpph.getUrl()
 
-    def pull_dsc(self):
-        "Retrieve dscfile and parse"
-        if self._dsc_source:
-            parsed = urlparse(self._dsc_source)
-            if parsed.scheme == '':
-                self._dsc_source = 'file://' + os.path.abspath(self._dsc_source)
-                parsed = urlparse(self._dsc_source)
-            self._download_dsc(self._dsc_source)
-        else:
-            for url in self._source_urls(self.dsc_name):
-                dlerr = None
-                try:
-                    self._download_dsc(url)
-                    break
-                except DownloadError as e:
-                    dlerr = e
-            else:
-                raise dlerr
-        self._check_dsc()
-
-    def _download_dsc(self, url):
-        "Download specified dscfile and parse"
-        parsed = urlparse(url)
-        if parsed.scheme == 'file':
-            Logger.debug("Using dsc file '%s'" % parsed.path)
-            with open(parsed.path, 'rb') as f:
-                body = f.read()
-        else:
-            try:
-                Logger.debug("Trying dsc url '%s'" % url)
-                response, body = httplib2.Http().request(url)
-            except httplib2.HttpLib2Error as e:
-                raise DownloadError(e)
-            if response.status != 200:
-                raise DownloadError("%s: %s %s" % (url, response.status,
-                                                   response.reason))
-        self._dsc = Dsc(body)
-
-    def _check_dsc(self):
-        "Check that the dsc matches what we are expecting"
-        assert self._dsc is not None
-        self.source = self.dsc['Source']
-        self._version = Version(self.dsc['Version'])
-
-        valid = False
-        no_pub_key = False
-        message = None
-        gpg_info = None
+    def _check_dsc_signature(self):
+        "Check that the dsc signature matches what we are expecting"
         try:
             gpg_info = self.dsc.get_gpg_info((
                 '/usr/share/keyrings/debian-keyring.gpg',
                 '/usr/share/keyrings/debian-maintainers.gpg',
             ))
-            valid = gpg_info.valid()
         except IOError:
-            message = ('Signature on %s could not be verified, install '
-                       'debian-keyring' % self.dsc_name)
-        if message is None:
-            if valid:
-                message = 'Valid signature'
-            else:
-                message = ('Signature on %s could not be verified'
-                           % self.dsc_name)
-        if gpg_info is not None:
+            Logger.debug('Signature on %s could not be verified, install '
+                         'debian-keyring' % self.dsc_name)
+            return
+        if gpg_info.valid():
             if 'GOODSIG' in gpg_info:
-                message = ('Good signature by %s (0x%s)'
-                           % (gpg_info['GOODSIG'][1], gpg_info['GOODSIG'][0]))
+                Logger.info('Good signature by %s (0x%s)'
+                            % (gpg_info['GOODSIG'][1], gpg_info['GOODSIG'][0]))
             elif 'VALIDSIG' in gpg_info:
-                message = 'Valid signature by 0x%s' % gpg_info['VALIDSIG'][0]
-            elif 'NO_PUBKEY' in gpg_info:
-                no_pub_key = True
-                message = 'Public key not found, could not verify signature'
-        if self._verify_signature:
-            if valid:
-                Logger.info(message)
-            elif no_pub_key:
-                Logger.warning(message)
+                Logger.info('Valid signature by 0x%s' % gpg_info['VALIDSIG'][0])
             else:
-                Logger.error(message)
-                raise DownloadError(message)
+                Logger.info('Valid signature')
+        elif 'NO_PUBKEY' in gpg_info:
+            Logger.warning('Public key not found, could not verify signature')
+        elif 'NODATA' in gpg_info:
+            Logger.warning('Package is not signed')
         else:
-            Logger.debug(message)
+            Logger.warning('Signature on %s could not be verified' % self.dsc_name)
 
-    def _write_dsc(self):
-        "Write dsc file to workdir"
-        if self._dsc is None:
-            self.pull_dsc()
-        with open(self.dsc_pathname, 'wb') as f:
-            f.write(self.dsc.raw_text)
-
-    def _verify_file(self, pathname, dscverify=False, sha1sum=False, sha256sum=False, size=0):
+    def _verify_file(self, pathname, dscverify=False, sha1sum=None, sha256sum=None, size=0):
         if not os.path.exists(pathname):
             return False
-        if size and size != os.path.getsize(pathname):
-            return False
         if dscverify and not self.dsc.verify_file(pathname):
             return False
-        if sha1sum and not verify_file_checksum(pathname, 'SHA1', sha1sum, size):
-            return False
-        if sha256sum and not verify_file_checksum(pathname, 'SHA256', sha256sum, size):
+        checksums = {}
+        if sha1sum:
+            checksums['SHA1'] = sha1sum
+        if sha256sum:
+            checksums['SHA256'] = sha256sum
+        if not verify_file_checksums(pathname, checksums, size):
             return False
         return True
 
-    def _download_file(self, url, filename, size, dscverify=False, sha1sum=None, sha256sum=None):
+    def _download_file(self, url, filename, size=0, dscverify=False, sha1sum=None, sha256sum=None):
         "Download url to filename in workdir."
         pathname = os.path.join(self.workdir, filename)
 
         if self._verify_file(pathname, dscverify, sha1sum, sha256sum, size):
-            Logger.debug('Using existing file %s', filename)
+            Logger.info('Using existing file %s', filename)
             return True
 
-        if urlparse(url).scheme in ["", "file"]:
-            frompath = os.path.abspath(urlparse(url).path)
-            if frompath == pathname:
-                Logger.info("Using %s" % pathname)
-            else:
-                Logger.info("Copying %s from %s" % (filename, frompath))
-                shutil.copyfile(frompath, pathname)
-        else:
-            download(url, pathname, size)
+        download(url, pathname, size)
 
         return self._verify_file(pathname, dscverify, sha1sum, sha256sum, size)
 
+    def _download_file_from_urls(self, urls, filename, size=0, dscverify=False,
+                                 sha1sum=None, sha256sum=None):
+        "Try to download a file from a list of urls."
+        for url in urls:
+            try:
+                if self._download_file(url, filename, size, dscverify=dscverify,
+                                       sha1sum=sha1sum, sha256sum=sha256sum):
+                    return
+            except HTTPError as e:
+                if e.code == 404:
+                    # It's ok if the file isn't found, just try the next url
+                    Logger.debug("File not found at %s" % url)
+                else:
+                    Logger.error('HTTP Error %i: %s', e.code, str(e))
+            except URLError as e:
+                Logger.error('URL Error: %s', e.reason)
+        raise DownloadError('Failed to download %s' % filename)
+
+    def pull_dsc(self):
+        "Pull dscfile into workdir"
+        urls = self._source_urls(self.dsc_name)
+        self._download_file_from_urls(urls, self.dsc_name)
+
     def pull(self):
         "Pull into workdir"
-        self._write_dsc()
         for entry in self.dsc['Files']:
             name = entry['name']
-            for url in self._source_urls(name):
-                try:
-                    if self._download_file(url, name, int(entry['size']), dscverify=True):
-                        break
-                except HTTPError as e:
-                    # It's ok if the file isn't found; we try multiple places to download
-                    if e.code == 404:
-                        Logger.info("File not found at %s" % url)
-                        continue
-                    Logger.info('HTTP Error %i: %s', e.code, str(e))
-                except URLError as e:
-                    Logger.info('URL Error: %s', e.reason)
-            else:
-                raise DownloadError('File %s could not be found' % name)
+            urls = self._source_urls(name)
+            self._download_file_from_urls(urls, name, int(entry['size']), dscverify=True)
 
     def pull_binaries(self, arch=None, name=None, ext=None):
         """Pull binary debs into workdir.
@@ -466,34 +427,25 @@ class SourcePackage(object):
 
         Returns the number of files downloaded.
         """
-        total = 0
-
         Logger.debug("pull_binaries(arch=%s, name=%s, ext=%s)" % (arch, name, ext))
 
         if arch == 'all':
             arch = None
 
+        total = 0
         for bpph in self.lp_spph.getBinaries(arch=arch, name=name, ext=ext):
             fname = bpph.getFileName()
             fsha1 = bpph.binaryFileSha1(fname)
             fsha256 = bpph.binaryFileSha256(fname)
             fsize = bpph.binaryFileSize(fname)
-            for url in self._binary_urls(fname, bpph):
-                try:
-                    if self._download_file(url, fname, fsize,
-                                           sha1sum=fsha1, sha256sum=fsha256):
-                        total += 1
-                        break
-                except HTTPError as e:
-                    # It's ok if the file isn't found; we try multiple places to download
-                    if e.code == 404:
-                        Logger.info("File not found at %s" % url)
-                        continue
-                    Logger.info('HTTP Error %i: %s', e.code, str(e))
-                except URLError as e:
-                    Logger.info('URL Error: %s', e.reason)
-            else:
-                Logger.info("Could not download from any location: %s", fname)
+            urls = self._binary_urls(fname, bpph)
+            try:
+                self._download_file_from_urls(urls, fname, fsize,
+                                              sha1sum=fsha1, sha256sum=fsha256)
+                total += 1
+            except DownloadError as e:
+                # log/print the error, but continue to get the rest of the files
+                Logger.error(e)
         return total
 
     def verify(self):
@@ -565,8 +517,14 @@ class DebianSPPH(SourcePackagePublishingHistory):
 
 class DebianSourcePackage(SourcePackage):
     "Download / unpack a Debian source package"
-    distribution = 'debian'
-    spph_class = DebianSPPH
+
+    @property
+    def distribution(self):
+        return 'debian'
+
+    @property
+    def spph_class(self):
+        return DebianSPPH
 
     def __init__(self, *args, **kwargs):
         super(DebianSourcePackage, self).__init__(*args, **kwargs)
@@ -617,26 +575,6 @@ class DebianSourcePackage(SourcePackage):
         if name in self.snapshot_files:
             yield self.snapshot_files[name]
 
-    def pull_dsc(self):
-        "Retrieve dscfile and parse"
-        try:
-            super(DebianSourcePackage, self).pull_dsc()
-            return
-        except DownloadError:
-            pass
-
-        # Not all Debian Source packages get imported to LP
-        # (or the importer could be lagging)
-        for url in self._source_urls(self.dsc_name):
-            try:
-                self._download_dsc(url)
-            except DownloadError:
-                continue
-            break
-        else:
-            raise DownloadError('dsc could not be found anywhere')
-        self._check_dsc()
-
     # Local methods:
     @property
     def snapshot_package(self):
@@ -674,7 +612,10 @@ class DebianSourcePackage(SourcePackage):
 
 class UbuntuSourcePackage(SourcePackage):
     "Download / unpack an Ubuntu source package"
-    distribution = 'ubuntu'
+
+    @property
+    def distribution(self):
+        return 'ubuntu'
 
 
 class PersonalPackageArchiveSourcePackage(UbuntuSourcePackage):
@@ -686,89 +627,206 @@ class PersonalPackageArchiveSourcePackage(UbuntuSourcePackage):
         if len(ppa) != 2:
             raise ValueError('Invalid PPA value "%s",'
                              'must be "<USER>/<PPA>"' % kwargs['ppa'])
-        self._set_ppa(ppa[0], ppa[1])
+        self._teamname = ppa[0]
+        self._ppaname = ppa[1]
         self.masters = []
 
-    def getArchive(self):
-        if not self._ppa:
-            try:
-                self._team = PersonTeam.fetch(self._ppateam)
-            except KeyError:
-                raise ValueError('No user/team "%s" found on Launchpad' % self._ppateam)
-            self._ppa = self._team.getPPAByName(self._ppaname)
-            Logger.debug('Using PPA %s' % self._ppa.web_link)
-        return self._ppa
+    @functools.cached_property
+    def team(self):
+        try:
+            return PersonTeam.fetch(self._teamname)
+        except KeyError:
+            raise ValueError(f"No user/team '{self._teamname}' found on Launchpad")
 
-    def _set_ppa(self, team, name):
-        self._ppateam = team
-        self._ppaname = name
-        self._team = None
-        self._ppa = None
+    @functools.lru_cache
+    def getArchive(self):
+        ppa = self.team.getPPAByName(self._ppaname)
+        Logger.debug(f"Using PPA '{ppa.web_link}'")
+        return ppa
 
 
 class UbuntuCloudArchiveSourcePackage(PersonalPackageArchiveSourcePackage):
     "Download / unpack an Ubuntu Cloud Archive source package"
-    _ppateam = 'ubuntu-cloud-archive'
-    _ppas = None
+    TEAM = 'ubuntu-cloud-archive'
+    PROJECT = 'cloud-archive'
+    VALID_POCKETS = ['updates', 'proposed', 'staging']
 
     def __init__(self, *args, **kwargs):
-        series = kwargs.pop('series', None)
-        check_all_series = series is None
-        if not series:
-            series = UbuntuCloudArchiveSourcePackage.getDevelSeries()
-        kwargs['ppa'] = ('%s/%s-staging' %
-                         (UbuntuCloudArchiveSourcePackage._ppateam, series))
+        # Need to determine actual series/pocket ppa now, as it affects getArchive()
+        (series, pocket) = self._findReleaseAndPocketForPackage(kwargs.get('series'),
+                                                                kwargs.get('pocket'),
+                                                                kwargs.get('package'),
+                                                                kwargs.get('version'))
+        # Drop series/pocket from kwargs, as UCA handles them completely different and we
+        # don't want to pass them up to the superclass
+        kwargs.pop('series', None)
+        orig_pocket = kwargs.pop('pocket', None)
+        if orig_pocket and orig_pocket != pocket and pocket == 'staging':
+            Logger.info(f"Ubuntu Cloud Archive release '{series}' pocket '{orig_pocket}'"
+                        " PPA is not public, using 'staging' pocket instead")
+
+        kwargs['ppa'] = f"{self.TEAM}/{series}-{pocket}"
         super(UbuntuCloudArchiveSourcePackage, self).__init__(*args, **kwargs)
-        self._uca_release = series
-        self._check_all_series = check_all_series
-        self.masters = ["http://ubuntu-cloud.archive.canonical.com/ubuntu/"]
+
+        if pocket == 'staging':
+            # Don't bother with the archive; just get directly from the staging ppa, since
+            # none of the binaries from the archive will match the staging checksums
+            self.masters = []
+        else:
+            self.masters = ["http://ubuntu-cloud.archive.canonical.com/ubuntu/"]
+
+    def pull_binaries(self, arch=None, name=None, ext=None):
+        """Pull binary debs into workdir.
+
+        This is only a wrapper around the superclass method, to log warning if
+        pulling binaries when using a 'staging' ppa, since the published binaries
+        will not match the 'staging' binaries.
+        """
+        if self._ppaname.endswith('-staging'):
+            Logger.warning("Binaries from 'staging' pocket will not match published binaries; "
+                           "see https://bugs.launchpad.net/cloud-archive/+bug/1649979")
+        return super(UbuntuCloudArchiveSourcePackage, self).pull_binaries(arch, name, ext)
 
     @classmethod
-    def getDevelSeries(cls):
-        return cls.ppas()[0]
+    @functools.lru_cache
+    def getUbuntuCloudArchiveProject(cls):
+        return Project(cls.PROJECT)
+
+    @classmethod
+    def getUbuntuCloudArchiveReleaseNames(cls):
+        """Get list of the main UCA release names
+
+        The list will be sorted in descending chronological order.
+        """
+        return [s.name for s in cls.getUbuntuCloudArchiveProject().series]
 
     @classmethod
     def ppas(cls):
-        if not cls._ppas:
-            ppas = PersonTeam.fetch(cls._ppateam).getPPAs().keys()
-            ppas = filter(lambda p: p.endswith('-staging'), ppas)
-            ppas = map(lambda p: p.rsplit('-', 1)[0], ppas)
-            ppas = sorted(ppas, reverse=True)
-            if not ppas:
-                raise SeriesNotFoundException('Internal Error: No UCA series found...?')
-            cls._ppas = ppas
-        return list(cls._ppas)
+        """DEPRECATED: use getUbuntuCloudArchiveReleaseNames()"""
+        return cls.getUbuntuCloudArchiveReleaseNames()
 
     @classmethod
     def isValidRelease(cls, release):
-        return release in cls.ppas()
+        return release in cls.getUbuntuCloudArchiveReleaseNames()
 
-    @property
-    def lp_spph(self):
-        "Return the LP Source Package Publishing History entry"
-        while True:
-            try:
-                return super(UbuntuCloudArchiveSourcePackage, self).lp_spph
-            except PackageNotFoundException as pnfe:
-                if self._check_all_series and self._set_next_release():
+    @classmethod
+    def getDevelSeries(cls):
+        """Get the current UCA devel release name"""
+        return cls.getUbuntuCloudArchiveReleaseNames()[0]
+
+    @classmethod
+    @functools.lru_cache
+    def getUbuntuCloudArchiveTeam(cls):
+        return PersonTeam.fetch(cls.TEAM)
+
+    @classmethod
+    @functools.lru_cache
+    def getUbuntuCloudArchivePPAs(cls, release=None, pocket=None):
+        """ Get sorted list of UCA ppa Archive objects
+
+        If release and/or pocket are specified, the list will be
+        filtered to return only matching ppa(s).
+
+        This list will only contain ppas relevant to UCA releases;
+        it will not contain 'other' ppas, e.g. 'cloud-tools-next'.
+        """
+        if not any((release, pocket)):
+            all_ppas = cls.getUbuntuCloudArchiveTeam().getPPAs()
+            ppas = []
+            for r in cls.getUbuntuCloudArchiveReleaseNames():
+                for p in cls.VALID_POCKETS:
+                    name = f"{r}-{p}"
+                    if name in all_ppas:
+                        ppas.append(all_ppas[name])
+            return ppas
+
+        # Use recursive call without params to get the lru-cached list of ppas returned above
+        ppas = cls.getUbuntuCloudArchivePPAs()
+        if release:
+            ppas = list(filter(lambda p: p.name.partition('-')[0] == release, ppas))
+        if pocket:
+            ppas = list(filter(lambda p: p.name.partition('-')[2] == pocket, ppas))
+        if not ppas:
+            rname = release or '*'
+            pname = pocket or '*'
+            raise SeriesNotFoundException(f"UCA release '{rname}-{pname}' not found")
+        return ppas
+
+    @classmethod
+    def parseReleaseAndPocket(cls, release):
+        """Parse the UCA release and pocket for the given string.
+
+        We allow 'release' to be specified as:
+           1. UCARELEASE
+           2. UCARELEASE-POCKET
+           3. UBUNTURELEASE-UCARELEASE
+           4. UBUNTURELEASE-UCARELEASE-POCKET
+           5. UBUNTURELEASE-POCKET/UCARELEASE
+        The UBUNTURELEASE is a standard Ubuntu release name, e.g. 'focal',
+        however it is NOT checked for validity.
+        The UCARELEASE is a standard Ubuntu Cloud Archive release name, e.g. 'train'.
+        The POCKET is limited to the standard pockets 'release', 'updates', or 'proposed',
+        or the special pocket 'staging'. The 'release' and 'updates' pockets are
+        equivalent (UCA has no 'release' pocket).
+
+        This will return a tuple of (release, pocket), or (release, None) if no
+        pocket was specified.
+
+        This will raise SeriesNotFoundException if the release and/or pocket were
+        not found.
+        """
+        release = release.lower().strip()
+
+        # Cases 1 and 2
+        PATTERN1 = r'^(?P<ucarelease>[a-z]+)(?:-(?P<pocket>[a-z]+))?$'
+        # Cases 3 and 4
+        PATTERN2 = r'^(?P<ubunturelease>[a-z]+)-(?P<ucarelease>[a-z]+)(?:-(?P<pocket>[a-z]+))?$'
+        # Case 5
+        PATTERN3 = r'^(?P<ubunturelease>[a-z]+)-(?P<pocket>[a-z]+)/(?P<ucarelease>[a-z]+)$'
+
+        for pattern in [PATTERN1, PATTERN2, PATTERN3]:
+            match = re.match(pattern, release)
+            if match:
+                r = match.group('ucarelease')
+                p = match.group('pocket')
+                # For UCA, there is no 'release' pocket, the default is 'updates'
+                if p and p == 'release':
+                    Logger.warning("Ubuntu Cloud Archive does not use 'release' pocket,"
+                                   " using 'updates' instead")
+                    p = 'updates'
+                if (cls.isValidRelease(r) and (not p or p in cls.VALID_POCKETS)):
+                    Logger.debug(f"Using Ubuntu Cloud Archive release '{r}'")
+                    return (r, p)
+        raise SeriesNotFoundException(f"Ubuntu Cloud Archive release '{release}' not found")
+
+    @classmethod
+    def _findReleaseAndPocketForPackage(cls, release, pocket, package, version):
+        if release and release not in cls.getUbuntuCloudArchiveReleaseNames():
+            raise SeriesNotFoundException(f"Ubuntu Cloud Archive release '{release}' not found")
+        if pocket and pocket not in cls.VALID_POCKETS:
+            raise PocketDoesNotExistError(f"Ubuntu Cloud Archive pocket '{pocket}' is invalid")
+        DEFAULT = tuple(cls.getUbuntuCloudArchivePPAs(release=release or cls.getDevelSeries())[0]
+                        .name.split('-', maxsplit=1))
+        if not package:
+            # not much we can do without a package name
+            return DEFAULT
+        checked_pocket = False
+        for ppa in cls.getUbuntuCloudArchivePPAs(release=release):
+            if pocket and pocket != ppa.name.partition('-')[2]:
+                # If pocket is specified, only look at the requested pocket, or 'later'
+                # This allows using the 'staging' pocket for old releases that do not
+                # provide any 'updates' or 'proposed' pockets
+                if not checked_pocket:
                     continue
-                raise pnfe
-
-    def _set_next_release(self):
-        ppas = UbuntuCloudArchiveSourcePackage.ppas()
-        try:
-            r = ppas[ppas.index(self._uca_release) + 1]
-        except IndexError:
-            return False
-        self._uca_release = r
-        self._set_ppa(UbuntuCloudArchiveSourcePackage._ppateam, '%s-staging' % r)
-        return True
-
-    def getArchive(self):
-        try:
-            return super(UbuntuCloudArchiveSourcePackage, self).getArchive()
-        except ValueError:
-            raise SeriesNotFoundException('UCA release {} not found.'.format(self._uca_release))
+            checked_pocket = True
+            params = {'exact_match': True, 'source_name': package}
+            if version:
+                params['version'] = version
+            if ppa.getPublishedSources(**params):
+                (r, _, p) = ppa.name.partition('-')
+                return (r, p)
+        # package/version not found in any ppa
+        return DEFAULT
 
 
 class _WebJSON(object):
