@@ -28,7 +28,9 @@ import locale
 import os
 import shutil
 import sys
+import tempfile
 
+from contextlib import suppress
 from subprocess import check_output, CalledProcessError
 from urllib.parse import urlparse
 from urllib.request import urlopen
@@ -199,7 +201,20 @@ def codename_to_distribution(codename):
             return distro
 
 
-def verify_file_checksum(pathname, alg, checksum, size=0):
+def verify_file_checksums(pathname, checksums={}, size=0):
+    """ verify checksums of file
+
+    Any failure will log an error.
+
+    pathname: str
+        full path to file
+    checksums: dict
+        keys are alg name, values are expected checksum
+    size: int
+        size of file, if known
+
+    Returns True if all checks pass, False otherwise
+    """
     if not os.path.isfile(pathname):
         Logger.error('File not found: %s', pathname)
         return False
@@ -208,39 +223,79 @@ def verify_file_checksum(pathname, alg, checksum, size=0):
         Logger.error('File %s incorrect size, got %s expected %s',
                      filename, os.path.getsize(pathname), size)
         return False
-    h = hashlib.new(alg)
-    with open(pathname, 'rb') as f:
-        while True:
-            block = f.read(h.block_size)
-            if len(block) == 0:
-                break
-            h.update(block)
-    match = h.hexdigest() == checksum
-    if match:
-        Logger.debug('File %s checksum (%s) verified: %s',
-                     filename, alg, checksum)
-    else:
-        Logger.error('File %s checksum (%s) mismatch: got %s expected %s',
-                     filename, alg, h.hexdigest(), checksum)
-    return match
+
+    for (alg, checksum) in checksums.items():
+        h = hashlib.new(alg)
+        with open(pathname, 'rb') as f:
+            while True:
+                block = f.read(h.block_size)
+                if len(block) == 0:
+                    break
+                h.update(block)
+        match = h.hexdigest() == checksum
+        if match:
+            Logger.debug('File %s checksum (%s) verified: %s',
+                         filename, alg, checksum)
+        else:
+            Logger.error('File %s checksum (%s) mismatch: got %s expected %s',
+                         filename, alg, h.hexdigest(), checksum)
+            return False
+    return True
 
 
-def download(src, dst=None, size=0):
-    """ download/copy a file/url to local file """
+def verify_file_checksum(pathname, alg, checksum, size=0):
+    """ verify checksum of file
+
+    pathname: str
+        full path to file
+    alg: str
+        name of checksum alg
+    checksum: str
+        expected checksum
+    size: int
+        size of file, if known
+
+    Returns True if all checks pass, False otherwise
+    """
+    return verify_file_checksums(pathname, {alg: checksum}, size)
+
+
+def download(src, dst, size=0):
+    """ download/copy a file/url to local file
+
+    src: str
+        Source to copy from (file path or url)
+    dst: str
+        Destination dir or filename
+    size: int
+        Size of source, if known
+
+    This calls urllib.request.urlopen() so it may raise the same
+    exceptions as that method (URLError or HTTPError)
+    """
+    if not urlparse(src).scheme:
+        src = 'file://%s' % os.path.abspath(os.path.expanduser(src))
+    dst = os.path.abspath(os.path.expanduser(dst))
+
     filename = os.path.basename(urlparse(src).path)
 
-    if not dst:
-        dst = filename
+    if os.path.isdir(dst):
+        dst = os.path.join(dst, filename)
+
+    if urlparse(src).scheme == 'file':
+        srcfile = urlparse(src).path
+        if os.path.exists(srcfile) and os.path.exists(dst):
+            if os.path.samefile(srcfile, dst):
+                Logger.info(f"Using existing file {dst}")
+                return
 
     with urlopen(src) as fsrc, open(dst, 'wb') as fdst:
         url = fsrc.geturl()
         Logger.debug(f"Using URL: {url}")
 
         if not size:
-            try:
+            with suppress(AttributeError, TypeError, ValueError):
                 size = int(fsrc.info().get('Content-Length'))
-            except (AttributeError, TypeError, ValueError):
-                pass
 
         hostname = urlparse(url).hostname
         sizemb = ' (%0.3f MiB)' % (size / 1024.0 / 1024) if size else ''
@@ -277,3 +332,20 @@ def download(src, dst=None, size=0):
                 Logger.error('Partial download: %0.3f MiB of %0.3f MiB' %
                              (downloaded / 1024.0 / 1024,
                               size / 1024.0 / 1024))
+
+
+def download_text(src):
+    """ return the text content of a downloaded file
+
+    src: str
+        Source to copy from (file path or url)
+
+    Raises the same exceptions as download()
+
+    Returns text content of downloaded file
+    """
+    with tempfile.TemporaryDirectory() as d:
+        dst = os.path.join(d, 'dst')
+        download(src, dst)
+        with open(dst) as f:
+            return f.read()
